import decimal
import json as _json
import os
from pathlib import Path
from typing import Any, Callable

import logging
import numpy as np
import pandas as pd
import yaml

# TODO(roulbac): Remove these imports once the nodes using them have been
# tested and updated to use huggingface_hub directly
from huggingface_hub import hf_hub_download
from huggingface_hub import snapshot_download

from flojoy.connection_manager import DeviceConnectionManager
from .dao import Dao
from .config import FlojoyConfig, logger
from .node_init import NodeInit, NodeInitService

import base64
from .CONSTANTS import FLOJOY_DIR, FLOJOY_CACHE_DIR, CREDENTIAL_FILE


__all__ = [
    "get_credentials",
    "hf_hub_download",
    "snapshot_download",
    "hf_hub_download",
    "snapshot_download",
    "clear_flojoy_memory",
]


# # package result
# def package_result(result: dict | None, fn: str, node_id: str, jobset_id: str) -> dict:
#     return {
#         "NODE_RESULTS": {
#             "cmd": fn,
#             "id": node_id,
#             "result": result,
#         },
#         "jobsetId": jobset_id,
#     }

# # package error
# def package_error(func_name, node_id, jobset_id, error):
#     return {
#         "SYSTEM_STATUS": f"Failed to run: {func_name}",
#         "FAILED_NODES": {node_id: str(error)},
#         "jobsetId": jobset_id,
#     }


# Make as a function to mock at test-time
def get_hf_hub_cache_path() -> str:
    """Returns the path to the HuggingFace home directory (HF_HOME) within the Flojoy cache directory
    This is used to cache huggingface artifacts within the Flojoy cache directory."""
    return os.path.join(FLOJOY_CACHE_DIR, "cache", "huggingface")


def set_offline():
    """
    Sets the is_offline flag to True, which means that results will not be sent to the backend via HTTP.
    Mainly used for precompilation
    """
    FlojoyConfig.get_instance().is_offline = True


def set_online():
    """
    Sets the is_offline flag to False, which means that results will be sent to the backend via HTTP.
    """
    FlojoyConfig.get_instance().is_offline = False


def set_debug_on():
    """
    Sets the print_on flag to True, which means that the print statements will be executed.
    """
    logger.setLevel(logging.DEBUG)


def set_debug_off():
    """
    Sets the print_on flag to False, which means that the print statements will not be executed.
    """
    logger.setLevel(logging.INFO)


def clear_flojoy_memory():
    Dao.get_instance().clear_job_results()
    Dao.get_instance().clear_small_memory()
    Dao.get_instance().clear_node_init_containers()
    DeviceConnectionManager.clear()


class PlotlyJSONEncoder(_json.JSONEncoder):
    """
    Meant to be passed as the `cls` kwarg to json.dumps(obj, cls=..)
    See PlotlyJSONEncoder.default for more implementation information.
    Additionally, this encoder overrides nan functionality so that 'Inf',
    'NaN' and '-Inf' encode to 'null'. Which is stricter JSON than the Python
    version.
    """

    def coerce_to_strict(self, const: Any):
        """
        This is used to ultimately *encode* into strict JSON, see `encode`
        """
        # before python 2.7, 'true', 'false', 'null', were include here.
        if const in ("Infinity", "-Infinity", "NaN"):
            return None
        else:
            return const

    def encode(self, o: Any):
        """
        Load and then dump the result using parse_constant kwarg
        Note that setting invalid separators will cause a failure at this step.
        """
        # this will raise errors in a normal-expected way
        encoded_o = super(PlotlyJSONEncoder, self).encode(o)
        # Brute force guessing whether NaN or Infinity values are in the string
        # We catch false positive cases (e.g. strings such as titles, labels etc.)
        # but this is ok since the intention is to skip the decoding / reencoding
        # step when it's completely safe

        if not ("NaN" in encoded_o or "Infinity" in encoded_o):
            return encoded_o

        # now:
        #    1. `loads` to switch Infinity, -Infinity, NaN to None
        #    2. `dumps` again so you get 'null' instead of extended JSON
        try:
            new_o = _json.loads(encoded_o, parse_constant=self.coerce_to_strict)
        except ValueError:
            # invalid separators will fail here. raise a helpful exception
            raise ValueError(
                "Encoding into strict JSON failed. Did you set the separators "
                "valid JSON separators?"
            )
        else:
            return _json.dumps(
                new_o,
                sort_keys=self.sort_keys,
                indent=self.indent,
                separators=(self.item_separator, self.key_separator),
            )

    def default(self, obj: dict[str, Any]):
        """
        Accept an object (of unknown type) and try to encode with priority:
        1. builtin:     user-defined objects
        2. sage:        sage math cloud
        3. pandas:      dataframes/series
        4. numpy:       ndarrays
        5. datetime:    time/datetime objects
        Each method throws a NotEncoded exception if it fails.
        The default method will only get hit if the object is not a type that
        is naturally encoded by json:
            Normal objects:
                dict                object
                list, tuple         array
                str, unicode        string
                int, long, float    number
                True                true
                False               false
                None                null
            Extended objects:
                float('nan')        'NaN'
                float('infinity')   'Infinity'
                float('-infinity')  '-Infinity'
        Therefore, we only anticipate either unknown iterables or values here.
        """
        # TODO: The ordering if these methods is *very* important. Is this OK?
        encoding_methods = (
            self.encode_as_plotly,
            self.encode_as_numpy,
            self.encode_as_pandas,
            self.encode_as_datetime,
            self.encode_as_date,
            self.encode_as_list,  # because some values have `tolist` do last.
            self.encode_as_decimal,
            self.encode_as_base64,
        )
        for encoding_method in encoding_methods:
            try:
                return encoding_method(obj)
            except NotEncodable:
                pass
        return _json.JSONEncoder.default(self, obj)

    @staticmethod
    def encode_as_base64(value: bytes):
        """Attempt to convert to base64."""
        try:
            return base64.b64encode(value).decode()
        except AttributeError:
            raise NotEncodable

    @staticmethod
    def encode_as_plotly(obj: dict[str, Any]):
        """Attempt to use a builtin `to_plotly_json` method."""
        try:
            return obj.to_plotly_json()
        except AttributeError:
            raise NotEncodable

    @staticmethod
    def encode_as_list(obj):
        """Attempt to use `tolist` method to convert to normal Python list."""
        if hasattr(obj, "tolist"):
            return obj.tolist()
        else:
            raise NotEncodable

    @staticmethod
    def encode_as_pandas(obj):
        """Attempt to convert pandas.NaT"""
        if not pd:
            raise NotEncodable
        if obj is pd.NaT:
            return None
        elif isinstance(obj, pd.DataFrame):
            return obj.to_json()
        else:
            raise NotEncodable

    @staticmethod
    def encode_as_numpy(obj):
        """Attempt to convert numpy.ma.core.masked"""
        if not np:
            raise NotEncodable

        if obj is np.ma.masked:
            return float("nan")
        elif isinstance(obj, np.ndarray) and obj.dtype.kind == "M":
            try:
                return np.datetime_as_string(obj).tolist()
            except TypeError:
                pass

        raise NotEncodable

    @staticmethod
    def encode_as_datetime(obj):
        """Convert datetime objects to iso-format strings"""
        try:
            return obj.isoformat()
        except AttributeError:
            raise NotEncodable

    @staticmethod
    def encode_as_date(obj):
        """Attempt to convert to utc-iso time string using date methods."""
        try:
            time_string = obj.isoformat()
        except AttributeError:
            raise NotEncodable
        else:
            return time_string  # iso_to_plotly_time_string(time_string)

    @staticmethod
    def encode_as_decimal(obj):
        """Attempt to encode decimal by converting it to float"""
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        else:
            raise NotEncodable


class NotEncodable(Exception):
    pass


def dump_str(result: Any, limit: int | None = None):
    result_str = str(result)
    return (
        result_str
        if limit is None or len(result_str) <= limit
        else result_str[:limit] + "..."
    )


def get_flojoy_root_dir() -> str:
    home = str(Path.home())
    path = os.path.join(home, ".flojoy/flojoy.yaml")
    stream = open(path, "r")
    yaml_dict = yaml.load(stream, Loader=yaml.FullLoader)
    root_dir = ""

    if isinstance(yaml_dict, str):
        root_dir = yaml_dict.split(":")[1]
    else:
        root_dir = yaml_dict["PATH"]

    return root_dir


def get_credentials() -> list[dict[str, str]]:
    home = str(Path.home())
    file_path = os.path.join(home, os.path.join(FLOJOY_DIR, CREDENTIAL_FILE))

    if not os.path.exists(file_path):
        logger.info(f"{file_path} not exists returning empty list...")
        return []

    with open(file_path, "r") as f:
        keys = f.read().strip().split(",")

    return [{"key": key, "value": ""} for key in keys]


def get_node_init_function(node_func: Callable) -> NodeInit:
    return NodeInitService().get_node_init_function(node_func)



from typing import Optional, Any, TypedDict
from .data_container import *  # noqa: F403
from .flojoy_python import *  # noqa: F403
from .job_result_builder import *  # noqa: F403
from .flojoy_instruction import *  # noqa: F403
from .plotly_utils import *  # noqa: F403
from .module_scraper import *  # noqa: F403
from .job_result_utils import *  # noqa: F403
from .utils import *  # noqa: F403
from .parameter_types import *  # noqa: F403
from .small_memory import *  # noqa: F403
from .flojoy_node_venv import *  # noqa: F403
from .job_service import *  # noqa: F403
from .node_init import *  # noqa: F403
from .node_preflight import *  # noqa: F403
from .config import *  # noqa: F403
from .flojoy_cloud import *  # noqa: F403
from .models import *  # noqa: F403

def flojoy(
    original_function: Callable[..., DataContainer | dict[str, Any] | TypedDict | None]  # noqa: F405
    | None = None,
    *,
    node_type: Optional[str] = None,
    deps: Optional[list[str]] = None,
    inject_node_metadata: bool = False,
    inject_connection: bool = False,
) -> Callable[..., DataContainer | dict[str, Any] | None]: ...  # noqa: F405



import os
import sys

KEY_WORKER_JOBS = "WORKER_JOBS"
KEY_ALL_JOBEST_IDS = "ALL_JOBSET_IDS"
FLOJOY_DIR = ".flojoy"
CREDENTIAL_FILE = "credentials.txt"
if sys.platform == "win32":
    FLOJOY_CACHE_DIR = os.path.realpath(os.path.join(os.environ["APPDATA"], FLOJOY_DIR))
else:
    FLOJOY_CACHE_DIR = os.path.realpath(os.path.join(os.environ["HOME"], FLOJOY_DIR))

KEYRING_KEY = "FLOJOY_KEYRING_KEY"



from .JobFeedback import JobFeedback

__all__ = ["JobFailure"]


class JobFailure(JobFeedback):
    def __init__(self, func_name, node_id, error, jobset_id):
        super().__init__(jobset_id)
        self.func_name = func_name
        self.node_id = node_id
        self.error = error



from threading import Lock

from tm_devices.drivers.pi.pi_device import PIDevice

from flojoy.parameter_types import (
    HardwareConnection,
    HardwareDevice,
    CameraDevice,
    CameraConnection,
    SerialDevice,
    SerialConnection,
    VisaDevice,
    VisaConnection,
    NIConnection,
    NIDMMDevice,
)
from tm_devices import DeviceManager
from tm_devices.helpers import PYVISA_PY_BACKEND
from typing import Any, Callable
from .config import logger

_connection_lock = Lock()


class DeviceConnectionManager:
    handles: dict[str | int, HardwareConnection] = {}
    tm = DeviceManager(verbose=False)
    tm.visa_library = PYVISA_PY_BACKEND

    @classmethod
    def register_connection(
        cls,
        device: HardwareDevice,
        connection: Any,
        cleanup: Callable[[Any], Any] | None = None,
    ):
        with _connection_lock:
            id = device.get_id()
            if id in cls.handles:
                raise ValueError(f"Connection with id {id} already exists")

            # Let tm_devices take care of any devices it created
            if isinstance(connection, PIDevice) and cleanup is None:
                cleanup = noop

            match device:
                case CameraDevice():
                    cls.handles[id] = CameraConnection(connection, cleanup=cleanup)
                case SerialDevice():
                    cls.handles[id] = SerialConnection(connection, cleanup=cleanup)
                case VisaDevice():
                    cls.handles[id] = VisaConnection(connection, cleanup=cleanup)
                case NIDMMDevice():
                    cls.handles[id] = NIConnection(connection, cleanup=cleanup)
                case HardwareDevice():
                    cls.handles[id] = HardwareConnection(connection, cleanup=cleanup)

    @classmethod
    def get_connection(cls, id: str | int) -> HardwareConnection:
        with _connection_lock:
            return cls.handles[id]

    @classmethod
    def clear(cls):
        cls.tm.remove_all_devices()
        logger.info("Cleaned up tm_devices DeviceManager")

        with _connection_lock:
            logger.info(f"Connections closed: {cls.handles}")
            cls.handles.clear()


def noop(_):
    return None



from .JobFeedback import JobFeedback

__all__ = ["JobSuccess"]


class JobSuccess(JobFeedback):
    def __init__(self, result, fn, node_id, jobset_id):
        super().__init__(jobset_id)
        self.result = result
        self.fn = fn
        self.node_id = node_id
        self.jobset_id = jobset_id



import sys
from keyrings.cryptfile.cryptfile import CryptFileKeyring
import keyring
from typing import Optional
from pathlib import Path
from .CONSTANTS import FLOJOY_DIR, CREDENTIAL_FILE, KEYRING_KEY
import os
from .config import logger

__all__ = [
    "get_env_var",
    "set_env_var",
    "delete_env_var",
    "get_flojoy_cloud_url",
]


def get_keyring():
    if sys.platform.lower() == "linux":
        kr = CryptFileKeyring()
        kr.keyring_key = KEYRING_KEY
        keyring.set_keyring(kr)
    return keyring


def get_env_var(key: str) -> Optional[str]:
    kr = get_keyring()
    return kr.get_password("flojoy", key)


def get_flojoy_cloud_url() -> str:
    # Key should be validated when assign, but just making sure
    cloud_url = get_env_var("FLOJOY_CLOUD_URL")
    if cloud_url is None:
        cloud_url = "https://api.flojoy.ai/"
    else:
        if not cloud_url.startswith("http"):
            cloud_url = f"https://{cloud_url}"
        if not cloud_url.endswith("/"):
            cloud_url = f"{cloud_url}/"

    return cloud_url


def set_env_var(key: str, value: str):
    kr = get_keyring()
    kr.set_password("flojoy", key, value)
    home = str(Path.home())
    file_path = os.path.join(home, os.path.join(FLOJOY_DIR, CREDENTIAL_FILE))

    if not os.path.exists(file_path):
        logger.info(f"{file_path} does not exist")
        with open(file_path, "w") as f:
            f.write(key)
        logger.info(f"Env var written to {file_path}")
        return

    logger.info(f"{file_path} exists, writing env to {file_path}")
    with open(file_path, "r") as f:
        keys = f.read().strip().split(",")
        if key not in keys:
            keys.append(key)

    with open(file_path, "w") as f:
        f.write(",".join(keys))


def delete_env_var(key: str):
    home = str(Path.home())
    file_path = os.path.join(home, os.path.join(FLOJOY_DIR, CREDENTIAL_FILE))

    if not os.path.exists(file_path):
        return

    with open(file_path, "r") as f:
        keys = f.read().strip().split(",")

    if key not in keys:
        return

    keys.remove(key)

    with open(file_path, "w") as f:
        f.write(",".join(keys))

    kr = get_keyring()
    kr.delete_password("flojoy", key)



"""
Tektronix RSA_API .h file Python Conversion
Author: Morgan Allison
Date created: 5/17
Date edited: 5/17
Windows 7 64-bit
RSA API version 3.9.0029
Python 3.6.0 64-bit (Anaconda 4.3.0)
Download Anaconda: http://continuum.io/downloads
Anaconda includes NumPy and MatPlotLib
Download the RSA_API: http://www.tek.com/model/rsa306-software
Download the RSA_API Documentation:
http://www.tek.com/spectrum-analyzer/rsa306-manual-6

YOU WILL NEED TO REFERENCE THE API DOCUMENTATION
"""

from ctypes import c_char_p, c_float, c_int16, c_int32, c_int, c_uint64
from ctypes import c_int64, c_double, c_bool, POINTER, c_uint32, c_uint16
from ctypes import c_wchar_p, Structure, c_uint8
from enum import Enum


class RSAError(Exception):
    pass


class ReturnStatus(Enum):
    noError = 0

    # Connection
    errorNotConnected = 101
    errorIncompatibleFirmware = 102
    errorBootLoaderNotRunning = 103
    errorTooManyBootLoadersConnected = 104
    errorRebootFailure = 105

    # POST
    errorPOSTFailureFPGALoad = 201
    errorPOSTFailureHiPower = 202
    errorPOSTFailureI2C = 203
    errorPOSTFailureGPIF = 204
    errorPOSTFailureUsbSpeed = 205
    errorPOSTDiagFailure = 206

    # General Msmt
    errorBufferAllocFailed = 301
    errorParameter = 302
    errorDataNotReady = 304

    # Spectrum
    errorParameterTraceLength = 1101
    errorMeasurementNotEnabled = 1102
    errorSpanIsLessThanRBW = 1103
    errorFrequencyOutOfRange = 1104

    # IF streaming
    errorStreamADCToDiskFileOpen = 1201
    errorStreamADCToDiskAlreadyStreaming = 1202
    errorStreamADCToDiskBadPath = 1203
    errorStreamADCToDiskThreadFailure = 1204
    errorStreamedFileInvalidHeader = 1205
    errorStreamedFileOpenFailure = 1206
    errorStreamingOperationNotSupported = 1207
    errorStreamingFastForwardTimeInvalid = 1208
    errorStreamingInvalidParameters = 1209
    errorStreamingEOF = 1210

    # IQ streaming
    errorIQStreamInvalidFileDataType = 1301
    errorIQStreamFileOpenFailed = 1302
    errorIQStreamBandwidthOutOfRange = 1303

    # -----------------
    # Internal errors
    # -----------------
    errorTimeout = 3001
    errorTransfer = 3002
    errorFileOpen = 3003
    errorFailed = 3004
    errorCRC = 3005
    errorChangeToFlashMode = 3006
    errorChangeToRunMode = 3007
    errorDSPLError = 3008
    errorLOLockFailure = 3009
    errorExternalReferenceNotEnabled = 3010
    errorLogFailure = 3011
    errorRegisterIO = 3012
    errorFileRead = 3013

    errorDisconnectedDeviceRemoved = 3101
    errorDisconnectedDeviceNodeChangedAndRemoved = 3102
    errorDisconnectedTimeoutWaitingForADcData = 3103
    errorDisconnectedIOBeginTransfer = 3104
    errorOperationNotSupportedInSimMode = 3015

    errorFPGAConfigureFailure = 3201
    errorCalCWNormFailure = 3202
    errorSystemAppDataDirectory = 3203
    errorFileCreateMRU = 3204
    errorDeleteUnsuitableCachePath = 3205
    errorUnableToSetFilePermissions = 3206
    errorCreateCachePath = 3207
    errorCreateCachePathBoost = 3208
    errorCreateCachePathStd = 3209
    errorCreateCachePathGen = 3210
    errorBufferLengthTooSmall = 3211
    errorRemoveCachePath = 3212
    errorGetCachingDirectoryBoost = 3213
    errorGetCachingDirectoryStd = 3214
    errorGetCachingDirectoryGen = 3215
    errorInconsistentFileSystem = 3216

    errorWriteCalConfigHeader = 3301
    errorWriteCalConfigData = 3302
    errorReadCalConfigHeader = 3303
    errorReadCalConfigData = 3304
    errorEraseCalConfig = 3305
    errorCalConfigFileSize = 3306
    errorInvalidCalibConstantFileFormat = 3307
    errorMismatchCalibConstantsSize = 3308
    errorCalConfigInvalid = 3309

    # flash
    errorFlashFileSystemUnexpectedSize = (3401,)
    errorFlashFileSystemNotMounted = 3402
    errorFlashFileSystemOutOfRange = 3403
    errorFlashFileSystemIndexNotFound = 3404
    errorFlashFileSystemReadErrorCRC = 3405
    errorFlashFileSystemReadFileMissing = 3406
    errorFlashFileSystemCreateCacheIndex = 3407
    errorFlashFileSystemCreateCachedDataFile = 3408
    errorFlashFileSystemUnsupportedFileSize = 3409
    errorFlashFileSystemInsufficentSpace = 3410
    errorFlashFileSystemInconsistentState = 3411
    errorFlashFileSystemTooManyFiles = 3412
    errorFlashFileSystemImportFileNotFound = 3413
    errorFlashFileSystemImportFileReadError = 3414
    errorFlashFileSystemImportFileError = 3415
    errorFlashFileSystemFileNotFoundError = 3416
    errorFlashFileSystemReadBufferTooSmall = 3417
    errorFlashWriteFailure = 3418
    errorFlashReadFailure = 3419
    errorFlashFileSystemBadArgument = 3420
    errorFlashFileSystemCreateFile = 3421

    # Aux monitoring
    errorMonitoringNotSupported = (3501,)
    errorAuxDataNotAvailable = 3502

    # battery
    errorBatteryCommFailure = 3601
    errorBatteryChargerCommFailure = 3602
    errorBatteryNotPresent = 3603

    # EST
    errorESTOutputPathFile = 3701
    errorESTPathNotDirectory = 3702
    errorESTPathDoesntExist = 3703
    errorESTUnableToOpenLog = 3704
    errorESTUnableToOpenLimits = 3705

    # Revision information
    errorRevisionDataNotFound = 3801

    # alignment
    error112MHzAlignmentSignalLevelTooLow = 3901
    error10MHzAlignmentSignalLevelTooLow = 3902
    errorInvalidCalConstant = 3903
    errorNormalizationCacheInvalid = 3904
    errorInvalidAlignmentCache = 3905

    # acq status
    errorADCOverrange = 9000  # must not change the location of these error codes without coordinating with MFG TEST
    errorOscUnlock = 9001

    errorNotSupported = 9901

    errorPlaceholder = 9999
    notImplemented = -1


class Cplx32(Structure):
    _fields_ = [("i", c_float), ("q", c_float)]


class CplxInt32(Structure):
    _fields_ = [("i", c_int32), ("q", c_int32)]


class CplxInt16(Structure):
    _fields_ = [("i", c_int16), ("q", c_int16)]


AcqDataStatus_ADC_OVERRANGE = 0x1
AcqDataStatus_REF_OSC_UNLOCK = 0x2
AcqDataStatus_LOW_SUPPLY_VOLTAGE = 0x10
AcqDataStatus_ADC_DATA_LOST = 0x20
AcqDataStatus_VALID_BITS_MASK = (
    AcqDataStatus_ADC_OVERRANGE
    or AcqDataStatus_REF_OSC_UNLOCK
    or AcqDataStatus_LOW_SUPPLY_VOLTAGE
    or AcqDataStatus_ADC_DATA_LOST
)


class AcqDataStatus:
    def __init__(self):
        self.adcOverrange = 0x1
        self.refFreqUnlock = 0x2
        self.lo1Unlock = 0x4
        self.lo2Unlock = 0x8
        self.lowSupplyVoltage = 0x10
        self.adcDataLost = 0x20
        self.event1pps = 0x40
        self.eventTrig1 = 0x80
        self.eventTrig2 = 0x100


DEVSRCH_MAX_NUM_DEVICES = 20
DEVSRCH_SERIAL_MAX_STRLEN = 100
DEVSRCH_TYPE_MAX_STRLEN = 20
DEVINFO_MAX_STRLEN = 100


class DEVICE_INFO(Structure):
    _fields_ = [
        ("nomenclature", c_char_p),
        ("serialNum", c_char_p),
        ("apiVersion", c_char_p),
        ("fwVersion", c_char_p),
        ("fpgaVersion", c_char_p),
        ("hwVersion", c_char_p),
    ]


class TriggerMode:
    def __init__(self):
        self.freeRun = c_int(0)
        self.triggered = c_int(1)


TriggerMode = TriggerMode()


class TriggerSource:
    def __init__(self):
        self.TriggerSourceExternal = c_int(0)
        self.TriggerSourceIFPowerLevel = c_int(1)


TriggerSource = TriggerSource()


class TriggerTransition:
    def __init__(self):
        self.TriggerTransitionLH = c_int(1)
        self.TriggerTransitionHL = c_int(2)
        self.TriggerTransitionEither = c_int(3)


TriggerTransition = TriggerTransition()


DEVEVENT_OVERRANGE = c_int(0)
DEVEVENT_TRIGGER = c_int(1)
DEVEVENT_1PPS = c_int(2)


class RunMode:
    def __init__(self):
        self.stopped = c_int(0)
        self.running = c_int(1)


RunMode = RunMode()


IQBLK_STATUS_INPUT_OVERRANGE = 1 << 0
IQBLK_STATUS_FREQREF_UNLOCKED = 1 << 1
IQBLK_STATUS_ACQ_SYS_ERROR = 1 << 2
IQBLK_STATUS_DATA_XFER_ERROR = 1 << 3


class IQBLK_ACQINFO(Structure):
    _fields_ = [
        ("sample0Timestamp", c_uint64),
        ("triggerSampleIndex", c_uint64),
        ("triggerTimestamp", c_uint64),
        ("acqStatus", c_uint32),
    ]


class IQHeader(Structure):
    _fields_ = [
        ("acqDataStatus", c_uint16),
        ("acquisitionTimestamp", c_uint64),
        ("frameID", c_uint32),
        ("trigger1Index", c_uint16),
        ("trigger2Index", c_uint16),
        ("timeSyncIndex", c_uint16),
    ]


class SpectrumWindows:
    def __init__(self):
        self.SpectrumWindow_Kaiser = c_int(0)
        self.SpectrumWindow_Mil6dB = c_int(1)
        self.SpectrumWindow_BlackmanHarris = c_int(2)
        self.SpectrumWindow_Rectangle = c_int(3)
        self.SpectrumWindow_FlatTop = c_int(4)
        self.SpectrumWindow_Hann = c_int(5)


SpectrumWindows = SpectrumWindows()


class SpectrumTraces:
    def __init__(self):
        self.SpectrumTrace1 = c_int(0)
        self.SpectrumTrace2 = c_int(1)
        self.SpectrumTrace3 = c_int(2)


SpectrumTraces = SpectrumTraces()


class SpectrumDetectors:
    def __init__(self):
        self.SpectrumDetector_PosPeak = c_int(0)
        self.SpectrumDetector_NegPeak = c_int(1)
        self.SpectrumDetector_AverageVRMS = c_int(2)
        self.SpectrumDetector_Sample = c_int(3)


SpectrumDetectors = SpectrumDetectors()


class SpectrumVerticalUnits:
    def __init__(self):
        self.SpectrumVerticalUnit_dBm = c_int(0)
        self.SpectrumVerticalUnit_Watt = c_int(1)
        self.SpectrumVerticalUnit_Volt = c_int(2)
        self.SpectrumVerticalUnit_Amp = c_int(3)
        self.SpectrumVerticalUnit_dBmV = c_int(4)


SpectrumVerticalUnits = SpectrumVerticalUnits()


class Spectrum_Settings(Structure):
    _fields_ = [
        ("span", c_double),
        ("rbw", c_double),
        ("enableVBW", c_bool),
        ("vbw", c_double),
        ("traceLength", c_int),
        ("window", c_int),
        ("verticalUnit", c_int),
        ("actualStartFreq", c_double),
        ("actualStopFreq", c_double),
        ("actualFreqStepSize", c_double),
        ("actualRBW", c_double),
        ("actualVBW", c_double),
        ("actualNumIQSamples", c_double),
    ]


class Spectrum_Limits(Structure):
    _fields_ = [
        ("maxSpan", c_double),
        ("minSpan", c_double),
        ("maxRBW", c_double),
        ("minRBW", c_double),
        ("maxVBW", c_double),
        ("minVBW", c_double),
        ("maxTraceLength", c_int),
        ("minTraceLength", c_int),
    ]


class Spectrum_TraceInfo(Structure):
    _fields_ = [("timestamp", c_int64), ("acqDataStatus", c_uint16)]


class DPX_FrameBuffer(Structure):
    _fields_ = [
        ("fftPerFrame", c_int32),
        ("fftCount", c_int64),
        ("frameCount", c_int64),
        ("timestamp", c_double),
        ("acqDataStatus", c_uint32),
        ("minSigDuration", c_double),
        ("minSigDurOutOfRange", c_bool),
        ("spectrumBitmapWidth", c_int32),
        ("spectrumBitmapHeight", c_int32),
        ("spectrumBitmapSize", c_int32),
        ("spectrumTraceLength", c_int32),
        ("numSpectrumTraces", c_int32),
        ("spectrumEnabled", c_bool),
        ("spectrogramEnabled", c_bool),
        ("spectrumBitmap", POINTER(c_float)),
        ("spectrumTraces", POINTER(POINTER(c_float))),
        ("sogramBitmapWidth", c_int32),
        ("sogramBitmapHeight", c_int32),
        ("sogramBitmapSize", c_int32),
        ("sogramBitmapNumValidLines", c_int32),
        ("sogramBitmap", POINTER(c_uint8)),
        ("sogramBitmapTimestampArray", POINTER(c_double)),
        ("sogramBitmapContainTriggerArray", POINTER(c_double)),
    ]


class DPX_SogramSettingStruct(Structure):
    _fields_ = [
        ("bitmapWidth", c_int32),
        ("bitmapHeight", c_int32),
        ("sogramTraceLineTime", c_double),
        ("sogramBitmapLineTime", c_double),
    ]


class DPX_SettingStruct(Structure):
    _fields_ = [
        ("enableSpectrum", c_bool),
        ("enableSpectrogram", c_bool),
        ("bitmapWidth", c_int32),
        ("bitmapHeight", c_int32),
        ("traceLength", c_int32),
        ("decayFactor", c_float),
        ("actualRBW", c_double),
    ]


class TraceType:
    def __init__(self):
        self.TraceTypeAverage = c_int(0)
        self.TraceTypeMax = c_int(1)
        self.TraceTypeMaxHold = c_int(2)
        self.TraceTypeMin = c_int(3)
        self.TraceTypeMinHold = c_int(4)


TraceType = TraceType()


class VerticalUnitType:
    def __init__(self):
        self.VerticalUnit_dBm = c_int(0)
        self.VerticalUnit_Watt = c_int(1)
        self.VerticalUnit_Volt = c_int(2)
        self.VerticalUnit_Amp = c_int(3)


VerticalUnitType = VerticalUnitType()


DPX_TRACEIDX_1 = c_int(0)
DPX_TRACEIDX_2 = c_int(1)
DPX_TRACEIDX_3 = c_int(2)


class AudioDemodMode:
    def __init__(self):
        self.ADM_FM_8KHZ = c_int(0)
        self.ADM_FM_13KHZ = c_int(1)
        self.ADM_FM_75KHZ = c_int(2)
        self.ADM_FM_200KHZ = c_int(3)
        self.ADM_AM_8KHZ = c_int(4)
        self.ADM_NONE = c_int(5)  # internal use only


AudioDemodMode = AudioDemodMode()


class StreamingMode:
    def __init__(self):
        self.StreamingModeRaw = c_int(0)
        self.StreamingModeFormatted = c_int(1)


StreamingMode = StreamingMode()


class IQSOUTDEST:
    def __init__(self):
        self.IQSOD_CLIENT = c_int(0)
        self.IQSOD_FILE_TIQ = c_int(1)
        self.IQSOD_FILE_SIQ = c_int(2)
        self.IQSOD_FILE_SIQ_SPLIT = c_int(3)


IQSOUTDEST = IQSOUTDEST()


class IQSOUTDTYPE:
    def __init__(self):
        self.IQSODT_SINGLE = c_int(0)
        self.IQSODT_INT32 = c_int(1)
        self.IQSODT_INT16 = c_int(2)


IQSOUTDTYPE = IQSOUTDTYPE()


IQSSDFN_SUFFIX_INCRINDEX_MIN = c_int(0)
IQSSDFN_SUFFIX_TIMESTAMP = c_int(-1)
IQSSDFN_SUFFIX_NONE = c_int(-2)

IFSSDFN_SUFFIX_INCRINDEX_MIN = c_int(0)
IFSSDFN_SUFFIX_TIMESTAMP = c_int(-1)
IFSSDFN_SUFFIX_NONE = c_int(-2)

IQSTRM_STATUS_OVERRANGE = 1 << 0
IQSTRM_STATUS_XFER_DISCONTINUITY = 1 << 1
IQSTRM_STATUS_IBUFF75PCT = 1 << 2
IQSTRM_STATUS_IBUFFOVFLOW = 1 << 3
IQSTRM_STATUS_OBUFF75PCT = 1 << 4
IQSTRM_STATUS_OBUFFOVFLOW = 1 << 5
IQSTRM_STATUS_NONSTICKY_SHIFT = 0
IQSTRM_STATUS_STICKY_SHIFT = 16

IQSTRM_MAXTRIGGERS = 100


class IQSTRMIQINFO(Structure):
    _fields_ = [
        ("timestamp", c_uint64),
        ("triggerCount", c_int),
        ("triggerIndices", POINTER(c_int)),
        ("scaleFactor", c_double),
        ("acqStatus", c_uint32),
    ]


class IQSTREAM_File_Info(Structure):
    _fields_ = [
        ("numberSamples", c_uint64),
        ("sample0Timestamp", c_uint64),
        ("triggerSampleIndex", c_uint64),
        ("triggerTimestamp", c_uint64),
        ("acqStatus", c_uint32),
        ("filenames", c_wchar_p),
    ]


class GNSS_SATSYS:
    def __init__(self):
        self.GNSS_NOSYS = c_int(0)
        self.GNSS_GPS_GLONASS = c_int(1)
        self.GNSS_GPS_BEIDOU = c_int(2)
        self.GNSS_GPS = c_int(3)
        self.GNSS_GLONASS = c_int(4)
        self.GNSS_BEIDOU = c_int(5)


GNSS_SATSYS = GNSS_SATSYS()


class POWER_INFO(Structure):
    _fields_ = [
        ("externalPowerPresent", c_bool),
        ("batteryPresent", c_bool),
        ("batteryChargeLevel", c_double),
        ("batteryCharging", c_bool),
        ("batteryOverTemperature", c_bool),
        ("batteryHardwareError", c_bool),
    ]



from .data_container import *  # noqa: F403
from .flojoy_python import *  # noqa: F403
from .job_result_builder import *  # noqa: F403
from .flojoy_instruction import *  # noqa: F403
from .plotly_utils import *  # noqa: F403
from .module_scraper import *  # noqa: F403
from .job_result_utils import *  # noqa: F403
from .data_container import *  # noqa: F403
from .utils import *  # noqa: F403
from .parameter_types import *  # noqa: F403
from .small_memory import *  # noqa: F403
from .flojoy_node_venv import *  # noqa: F403
from .job_service import *  # noqa: F403
from .node_init import *  # noqa: F403
from .config import *  # noqa: F403
from .node_preflight import *  # noqa: F403
from .instruments import *  # noqa: F403
from .models import *  # noqa: F403
from .connection_manager import *  # noqa: F403
from .env_var import *  # noqa: F403



from flojoy.env_var import get_env_var
from typing import Any, NewType, Union, Callable
from abc import ABC


class HardwareDevice(ABC):
    # Some unique identifier for the hardware device,
    # For cameras, this is either the port it's connected to or the camera index
    # For serial devices, this is the connection port
    # For visa devices, this is the visa address
    _id: str | int

    def __init__(self, id: int | str) -> None:
        if id == "":
            raise ValueError(
                "No device selected, please select one using the parameter menu."
            )
        self._id = id

    def get_id(self):
        return self._id


class HardwareConnection(ABC):
    _handle: Any
    _cleanup: Callable[[Any], Any]

    def __init__(self, handle: Any, cleanup: Callable[[Any], Any]) -> None:
        self._handle = handle
        self._cleanup = cleanup

    def get_handle(self):
        return self._handle

    def __del__(self):
        self._cleanup(self._handle)


class CameraDevice(HardwareDevice):
    pass


class SerialDevice(HardwareDevice):
    def get_port(self):
        return str(self.get_id())


class VisaDevice(HardwareDevice):
    def get_address(self):
        return str(self.get_id())


class NIDAQmxDevice(HardwareDevice):
    def get_addresses(self):
        return str(self.get_id())


class NIDMMDevice(HardwareDevice):
    def get_address(self):
        return str(self.get_id())


class NIDevice(HardwareDevice):
    pass


class CameraConnection(HardwareConnection):
    def __init__(
        self, handle: Any, cleanup: Callable[[Any], Any] | None = None
    ) -> None:
        super().__init__(
            handle, (lambda conn: conn.release()) if cleanup is None else cleanup
        )

    def __del__(self):
        super().__del__()


class SerialConnection(HardwareConnection):
    def __init__(
        self, handle: Any, cleanup: Callable[[Any], Any] | None = None
    ) -> None:
        super().__init__(
            handle, (lambda conn: conn.close()) if cleanup is None else cleanup
        )

    def __del__(self):
        super().__del__()


class VisaConnection(HardwareConnection):
    def __init__(
        self, handle: Any, cleanup: Callable[[Any], Any] | None = None
    ) -> None:
        super().__init__(
            handle, (lambda conn: conn.close()) if cleanup is None else cleanup
        )

    def __del__(self):
        super().__del__()


class NIConnection(HardwareConnection):
    def __init__(
        self, handle: Any, cleanup: Callable[[Any], Any] | None = None
    ) -> None:
        super().__init__(
            handle, (lambda conn: conn.close()) if cleanup is None else cleanup
        )

    def __del__(self):
        super().__del__()


class NodeReference:
    """Node parameter type"""

    ref: str

    def __init__(self, ref: str) -> None:
        self.ref = ref

    def unwrap(self):
        return self.ref


class Array:
    """Node parameter type of `list[str | float | int]`"""

    ref: list[str | float | int]

    def __init__(self, ref: list[str | float | int]) -> None:
        self.ref = ref

    def unwrap(self):
        return self.ref


class File:
    """Node parameter type of str"""

    ref: str

    def __init__(self, ref: str) -> None:
        self.ref = ref

    def unwrap(self):
        return self.ref


class Directory:
    """Node parameter type of str"""

    ref: str

    def __init__(self, ref: str) -> None:
        self.ref = ref

    def unwrap(self):
        return self.ref


class Secret:
    """Node parameter type of str"""

    ref: str

    def __init__(self, ref: str) -> None:
        self.ref = ref

    def unwrap(self):
        return get_env_var(self.ref)


TextArea = NewType("TextArea", str)


def format_param_value(value: Any, value_type: str):
    match value_type:
        case "Array":
            s = str(value)
            parsed_value = parse_array(s, [str, float, int], "list[int | float | str]")
            return Array(parsed_value)
        case "list[str]":
            return parse_array(str(value), [str], "list[str]")
        case "list[float]":
            return parse_array(str(value), [float], "list[float]")
        case "list[int]":
            return parse_array(str(value), [int], "list[int]")
        case "select" | "str":
            return str(value)
        case "TextArea":
            return TextArea(value)
        case "Secret":
            return Secret(value)
        case "CameraDevice" | "CameraConnection":
            return (
                CameraDevice(int(value)) if value.isnumeric() else CameraDevice(value)
            )
        case "SerialDevice" | "SerialConnection":
            return SerialDevice(value)
        case "VisaDevice" | "VisaConnection":
            return VisaDevice(value)
        case "NIDAQmxDevice":
            return NIDAQmxDevice(value)
        case "NIDMMDevice" | "NIConnection":
            return NIDMMDevice(value)

    if value == "":
        return None

    match value_type:
        case "float":
            return float(value)
        case "int":
            return int(value)
        case "bool":
            return bool(value)
        case "NodeReference":
            return NodeReference(str(value))
        case "File":
            return File(str(value))
        case "Directory":
            return Directory(str(value))
        case _:
            print("hit default case", flush=True)
            return value


def parse_array(
    str_value: str, type_list: list[Any], param_type: str
) -> list[Union[int, float, str]]:
    if not str_value:
        return []

    val_list = [val.strip() for val in str_value.split(",")]
    # First try to cast into int, then float, then keep as string if all else fails
    for t in type_list:
        try:
            return list(map(t, val_list))
        except ValueError:
            continue

    raise ValueError(
        f"Couldn't parse list items with type {','.join([str(t) for t in type_list])}."
        + f"Value should be comma (',') separated {' | '.join([t.__name__ for t in type_list])} for parameter type {param_type}."
    )
